!*****************
! File Description:  
!   Full EC coupling whole-cell model
! Creator: Hoang-Trong Minh Tuan 
! 
! Rev: 0.1
! Update: 
!    0.9: use new param file structure (with RYR and LCC Markov models are in 2 separate files)
!    1.0: major revision (new algorithm) (March, 2nd, 2012 - Tuan)
!    2.0: major update Markov model algorithm (Tuan, 2013-Oct-31)
!  NOTE:
!    if segmentation falt occurs, make sure the correct version of ACML is being used
!*******************/

PROGRAM SparkCode_GPU
  USE precisioninfo
  USE makeKs_utility
  USE data_global
  USE data_ecc
  USE error_codes
  USE hdf5
  USE ecc_gpu_utility
  USE cudafor
  IMPLICIT NONE
#include "macro.h"

  INTEGER :: ierr
  INTEGER :: hdf_error
  INTEGER :: version_info(3)
  INTEGER(2) :: ic2, cmax2
  INTEGER(4) :: ic4, cmax4
  REAL :: crate2, crate4
  CHARACTER(100) :: argu
  ierr = SUCCESS

!!$  CALL GET_COMMAND_ARGUMENT(1, argu)
!!$  READ(argu, '(I10)') ierr
!!$  CALL GET_COMMAND_ARGUMENT(2, argu)
!!$  READ(argu, '(I10)') hdf_error
!!$  PRINT *, "<bin> Nchannl Nstate"
!!$  CALL export_Umat(ierr, hdf_error)
!!$  !!CALL export_Umat(49,2)
!!$  CALL state_breakdown(50, 20, ierr, hdf_error)
!!$  PRINT *, ierr, hdf_error
!!$  PRINT *, state_combine_single(ierr, hdf_error, 20)
!!$  stop
  
  CALL check_expiration()
  
!  CALL H5get_libversion_f(version_info(1), version_info(2), version_info(3), hdf_error)
!  WRITE(*,"(A i0 '.' i0 '.' i0)") "HDF5 version:", version_info

  CALL SYSTEM_CLOCK(ic2, crate2, cmax2)  !for benchmarking purpose

  ierr =  prog_get_info()
  IF (ierr .NE. SUCCESS) THEN
     CALL cleanup()
  END IF

  ierr = check_ecc_data()
  IF (ierr .NE. SUCCESS) THEN
     CALL cleanup()
  END IF
  
  ierr = init_GPU(gpuID)
  IF (ierr .NE. SUCCESS) THEN
     CALL cleanup()
  END IF

!!$  PRINT *, "g_bCa", find_bg_Ca_conductance()
!!$  stop
!!$  PRINT *, "v_ryr_T", find_v_ryr(0.22d0, V_myo_T, 20000, N_R)
!!$  stop
!!$  ierr = setup_RYR_2state_new(ryr_Krateconstant, N_L, mL, iU_L, &
!!$       ryr_stateFromTo, ryr_indxK, &
!!$       U_Lo, irow_L, ryr_StateSpace, ryr_maxNumNeighbors)

  
  ierr =  setup_RYR(ryr_KCal_act, ryr_Kfree, N_R, mR, iU_R, ryr_compKcal_act, ryr_compKfree, &
       ryr_indxKcal_act, ryr_indxKfree, &
       U_Ro, ryr_Ncol_Kcal_act, ryr_Ncol_Kfree, irow_R)
  IF (ierr .NE. SUCCESS) THEN
     CALL cleanup()
  END IF

  !CALL print_mat(ryr_compKcal_act, 1, irow_R, 0, ryr_Ncol_Kcal_act)
  !CALL print_mat(ryr_indxKcal_act, 1, irow_R, 0, ryr_Ncol_Kcal_act)
  !CALL print_mat(U_Ro, 1, irow_R, 0, ryr_Ncol_Kcal_act)
  !CALL print_mat(ryr_compKfree, 1, irow_R, 0, ryr_Ncol_Kfree)
  !STOP
  
#if _LCC_MODEL_ == _WILLIAMS2013 
  ierr = setup_LCC(lcc_Kv_act, lcc_Kv_inact, lcc_KCal_inact, lcc_Kfree, N_L, mL, iU_L, &
       lcc_compKv_act, lcc_compKv_inact, lcc_compKcal_inact, lcc_compKfree, &
       lcc_indxKv_act, lcc_indxKv_inact, lcc_indxKcal_inact, lcc_indxKfree, &
       U_Lo, lcc_Ncol_Kv_act, lcc_Ncol_Kv_inact, lcc_Ncol_Kcal_inact, lcc_Ncol_Kfree, irow_L)
!#elif _LCC_MODEL_ == _SUN2000 || _LCC_MODEL_ == _WILLIAMS2013_NEWALGORITHM || _LCC_MODEL_ == _MAHAJAN2008 || _LCC_MODEL_ == _MOROTTI2012
#else  
  ierr = setup_LCC_Sun2000(lcc_Krateconstant, N_L, mL, iU_L, &
       lcc_stateFromTo, lcc_indxK, &
       U_Lo, irow_L, lcc_StateSpace, lcc_maxNumNeighbors)
#endif

  !CALL print_mat(lcc_compKv_act, 1, irow_L, 0, lcc_Ncol_Kv_act)
  !CALL print_mat(lcc_indxKv_act, 1, irow_L, 0, lcc_Ncol_Kv_act)
  !CALL print_vect_T(U_Lo, irow_L)
  !STOP
  !PRINT *, lcc_Ncol_Kv_act, lcc_Ncol_Kv_inact, ryr_Ncol_Kfree
  !stop
  IF (ierr .NE. SUCCESS) THEN
     CALL cleanup()
  END IF

#if _LCC_MODEL_ == _WILLIAMS2013
  ierr = makeKs_nj(ryr_KCal_act, ryr_Kfree, mR, iU_R, ryr_initialstate, &
       lcc_Kv_act, lcc_Kv_inact, lcc_KCal_inact, lcc_Kfree, mL, iU_L, lcc_initialstate, &
       aKv1_nj, aKv2_nj, aKm1_nj, aKp1_nj,&
          aKm2_nj, aKp2_nj, U_Lo_nj, U_Ro_nj, pL_nj, pR_nj, irow_L_nj, irow_R_nj)
#endif
  IF (ierr .NE. SUCCESS) THEN
     CALL cleanup()
  END IF

#if _TEST_ORPHAN_CRU_ == _YES
  CALL makeKs_orphan_2(N_R, aKm2_orphan, aKp2_orphan, U_Ro_orphan, irow_R_orphan)
!!$  CALL print_mat(U_Ro_orphan, irow_R_orphan, 1)
!!$  PRINT *, U_Ro_orphan(1:irow_R_orphan)
!!$ STOP
#endif  

  ierr = init_Simulation()
  IF (ierr .NE. SUCCESS) THEN
     CALL cleanup()
  END IF

  CALL continue_Simulation()

  CALL finalize_Simulation()

  CALL SYSTEM_CLOCK(count=ic4)
  WRITE(log_string, *)  "the elapsed time (sec) is ", DBLE((ic4-ic2))/crate2
  CALL write2log(log_string)
  
  !free
  CALL program_cleanup()
  WRITE (log_string, *) "Program completed!!!" 
  CALL write2log(log_string)
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
CONTAINS
  SUBROUTINE program_cleanup()
    CALL cleanup_data_ecc()
    CALL gpu_cleanup()
    CALL cleanup()
  END SUBROUTINE program_cleanup
END PROGRAM SparkCode_GPU
 
